ONLY FORTH ALSO SIM16 DEFINITIONS

( eM24 Simulator, 21jun00cht )
( derived from P16.txt 06jun99cht)
( 02sep00cht, eliminate R, X, Y registers)
( SIM25.seq, ok for ef24 )
( SIM26, 29oct00cht, move to win32forth )
( sim24, 09nov00cht, for P24 )
( sim24b.f, 27nov00cht, align to sim26f.f with get and put)
( 30nov00cht, change to P24, 6 bit instrucitons )
( 03jan01cht, vector KEY EMIT to new map in kern24c )
( sim24f.f, 05jan01cht, increase return stack to 32 levels )
( sim16.f, 17jan01cht, adapted to p16b )

DECIMAL

31 CONSTANT SLIMIT       ( data stack depth )
31 CONSTANT RLIMIT       ( return stack depth )
$1FFF CONSTANT RANGE   ( size of memory array )
VARIABLE CLOCK          ( slot is in the last 3 bits )
VARIABLE (REGISTER)     ( where registers and stacks are )
VARIABLE BREAK          ( address of break point )

( On the rising edge of clock, copy TO array to FROM array. )
: REGISTER  (REGISTER) @ ;
: FROM  PAD (REGISTER) ! ;
: TO    PAD $180 + (REGISTER) ! ;

: P     REGISTER ;
: T     REGISTER 4 + ;
: A     REGISTER 12 + ;
: I     REGISTER 24 + ;
: I0    REGISTER 28 + ;
: I1    REGISTER 29 + ;
: I2    REGISTER 30 + ;
: I3    REGISTER 31 + ;
: I4    REGISTER 32 + ;
: RP    REGISTER 33 + ;
: SP    REGISTER 34 + ;
: RSTACK  RP C@ RLIMIT AND CELLS REGISTER + $40 + ;
: SSTACK  SP C@ SLIMIT AND CELLS REGISTER + $C0 + ;

: CYCLE TO P FROM P $180 CMOVE 1 CLOCK +! ;

: NEXT  CLOCK @ 3 OR CLOCK ! ;

: RPUSH ( d -- , push d on return stack )
        FROM RP C@ 1 + RLIMIT AND TO RP C! RSTACK ! ;

: RPOPP ( -- d , pop d from return stack )
        FROM RSTACK @ RP C@ 1 - RLIMIT AND TO RP C! ;

: SPUSH ( d -- , push d on data stack )
        FROM T @ SP C@ 1 + SLIMIT AND TO SP C! SSTACK ! T ! ;

: SPOPP ( -- d , pop d from data stack )
        FROM T @ SSTACK @ SP C@ 1 - SLIMIT AND TO SP C! T ! ;

: continue
        FROM P @ DUP 1+ TO RANGE AND P !
        RAM@ DUP I ! 2/
        32 /MOD SWAP I3 C!
        32 /MOD SWAP I2 C!
        31 AND I1 C!
        ;

: jmp   FROM I @ $7FF AND P @ $1800 AND + TO P ! NEXT ;
: ret   RPOPP TO RANGE AND P ! NEXT ;
: jz    FROM T @ $FFFF AND IF NEXT ELSE jmp THEN ;
: jnc   FROM T @ $10000 AND IF NEXT ELSE jmp THEN ;
: call  FROM P @ RPUSH jmp ;
: ld    FROM A @ RANGE AND RAM@ SPUSH ;
: ldp   ld FROM A @ 1+ TO A ! ;
: ldi   FROM P @ 1+ RANGE AND TO P !
        FROM P @ RANGE AND RAM@ SPUSH ;
: st    SPOPP FROM A @ RANGE AND RAM! ;
: stp   st FROM A @ 1+ TO A ! ;
: com   FROM T @ $FFFF AND $FFFF XOR TO T ! ;
: shr   FROM T @ DUP $8000 AND SWAP 2/ $FFFF AND OR TO T ! ;
: shl   FROM T @ 2* $1FFFF AND TO T ! ;
: addc  FROM T @ 1 AND
        IF SSTACK @ $FFFF AND T @ $FFFF AND + TO T ! THEN ;
: andd  SPOPP TO T @ AND $FFFF AND T ! ;
: xorr  SPOPP TO T @ XOR $FFFF AND T ! ;
: add   SPOPP $FFFF AND TO T @ $FFFF AND + TO T ! ;
: popr  RPOPP SPUSH ;
: pushs FROM T @ SPUSH ;
: lda   FROM A @ SPUSH ;
: pushr SPOPP RPUSH ;
: sta   SPOPP TO A ! ;
: pops  SPOPP DROP ;
: nop   NEXT ;
: get   KEY DUP $1B = ABORT" done"
        SPUSH ;
: put   SPOPP $7F AND EMIT ;

HEX

: execute ( code -- )
        DUP  0 = IF DROP jmp  EXIT THEN
        DUP  1 = IF DROP ret  EXIT THEN
        DUP  2 = IF DROP jz   EXIT THEN
        DUP  3 = IF DROP jnc  EXIT THEN
        DUP  4 = IF DROP call EXIT THEN
        DUP  6 = IF DROP get  EXIT THEN
        DUP  7 = IF DROP put  EXIT THEN
        DUP  9 = IF DROP ldp  EXIT THEN
        DUP 0A = IF DROP ldi  EXIT THEN
        DUP 0B = IF DROP ld   EXIT THEN
        DUP 0D = IF DROP stp  EXIT THEN
        DUP 0F = IF DROP st   EXIT THEN
        DUP 10 = IF DROP com  EXIT THEN
        DUP 11 = IF DROP shl  EXIT THEN
        DUP 12 = IF DROP shr  EXIT THEN
        DUP 13 = IF DROP addc EXIT THEN
        DUP 14 = IF DROP xorr EXIT THEN
        DUP 15 = IF DROP andd EXIT THEN
        DUP 17 = IF DROP add  EXIT THEN
        DUP 18 = IF DROP popr EXIT THEN
        DUP 19 = IF DROP lda  EXIT THEN
        DUP 1A = IF DROP pushs EXIT THEN
        DUP 1C = IF DROP pushr EXIT THEN
        DUP 1D = IF DROP sta  EXIT THEN
        DUP 1E = IF DROP nop  EXIT THEN
        DUP 1F = IF DROP pops EXIT THEN
        CR . ." illegal code" ABORT
        ;

: .sstack ." S:" T @ . SP C@
        SLIMIT 1+ 0 DO DUP SP C! SSTACK @ . 1- LOOP
        SLIMIT AND SP C! CR ;
: .rstack ." R:" RP C@
        RLIMIT 1+ 0 DO DUP RP C! RSTACK @ . 1- LOOP
        RLIMIT AND RP C! CR ;
: .registers ."  P=" P @ . ."  I=" I @ .
        ." I1=" I1 C@ . ." I2=" I2 C@ .
        ." I3=" I3 C@ . ." SP=" SP C@ . ." RP=" RP C@ . CR
        ." A=" A @ . CR ;
: S  CR ." CLOCK=" CLOCK @ . .registers
        .sstack .rstack ;

: sync  CLOCK @ 7 AND
        DUP 0 = IF   continue DROP EXIT THEN
        DUP 1 = IF   I1 C@ execute DROP EXIT
                THEN
        DUP 2 = IF   I2 C@ execute DROP EXIT
                THEN
        DUP 3 = IF   I3 C@ execute THEN
        DROP NEXT ;
: C     sync CYCLE S ;
: reset FROM P 300 ERASE 0 CLOCK ! ;
reset

: G     ( addr -- )
        CR ." Press any key to stop." CR
        BREAK !
        BEGIN sync P @ BREAK @ =
              IF CYCLE C EXIT
              ELSE CYCLE
              THEN
              KEY?
        UNTIL ;
: PUSH  ( d ) pushs TO T ! ;
: POP   pops ;

: D     P @ 1- FOUR FOUR ;
: M     SHOW ;
: RUN   CR ." Press ESC to stop." CR
        BEGIN C KEY 1B = UNTIL ;
: P     RANGE AND DUP FROM P ! TO P ! ;

: HELP  CR ." eM24 Simulator, copyright eMAST Technology, 2000"
        CR ." C: execute next cycle"
        CR ." S: show all registers"
        CR ." D: display next 8 words"
        CR ." addr M: display 128 words from addr"
        CR ." addr P: start execution at addr"
        CR ." addr G: run and stop at addr"
        CR ." RUN: execute, one key per cycle"
        CR ;

( patch KEY and EMIT to run eForth interactively )
ASM16
forth' KEY  >body forth@ ORG
   get @+ push ;'
forth' EMIT >body forth@ ORG
   put @+ push ;'
SIM16

